#!/usr/bin/env python
#
# PyQuarantine-Milter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQuarantine-Milter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQuarantineMilter.  If not, see <http://www.gnu.org/licenses/>.
#

import Milter
import argparse
import logging
import logging.handlers
import sys

import pyquarantine

from pyquarantine.version import __version__ as version

def main():
    "Run PyQuarantine-Milter."
    # parse command line
    parser = argparse.ArgumentParser(
        description="PyQuarantine milter daemon",
        formatter_class=lambda prog: argparse.HelpFormatter(
            prog, max_help_position=45, width=140))
    parser.add_argument(
        "-c", "--config",
        help="List of config files to read.",
        nargs="+",
        default=pyquarantine.QuarantineMilter.get_configfiles())
    parser.add_argument(
        "-s", "--socket",
        help="Socket used to communicate with the MTA.",
        required=True)
    parser.add_argument(
        "-d", "--debug",
        help="Log debugging messages.",
        action="store_true")
    parser.add_argument(
        "-t", "--test",
        help="Check configuration.",
        action="store_true")
    parser.add_argument(
        "-v", "--version",
        help="Print version.",
        action="version",
        version="%(prog)s ({})".format(version))
    args = parser.parse_args()

    # setup logging
    loglevel = logging.INFO
    logname = "pyquarantine-milter"
    syslog_name = logname
    if args.debug:
        loglevel = logging.DEBUG
        logname = "{}[%(name)s]".format(logname)
        syslog_name = "{}: [%(name)s] %(levelname)s".format(syslog_name)

    # set config files for milter class
    pyquarantine.QuarantineMilter.set_configfiles(args.config)
    root_logger = logging.getLogger()
    root_logger.setLevel(loglevel)

    # setup console log
    stdouthandler = logging.StreamHandler(sys.stdout)
    stdouthandler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(message)s".format(logname))
    stdouthandler.setFormatter(formatter)
    root_logger.addHandler(stdouthandler)
    logger = logging.getLogger(__name__)
    if args.test:
        try:
            pyquarantine.generate_milter_config(args.test)
            print("Configuration ok")
        except RuntimeError as e:
            logger.error(e)
            sys.exit(255)
        else:
            sys.exit(0)
    formatter = logging.Formatter(
        "%(asctime)s {}: [%(levelname)s] %(message)s".format(logname),
        datefmt="%Y-%m-%d %H:%M:%S")
    stdouthandler.setFormatter(formatter)

    # setup syslog
    sysloghandler = logging.handlers.SysLogHandler(
        address="/dev/log", facility=logging.handlers.SysLogHandler.LOG_MAIL)
    sysloghandler.setLevel(loglevel)
    formatter = logging.Formatter("{}: %(message)s".format(syslog_name))
    sysloghandler.setFormatter(formatter)
    root_logger.addHandler(sysloghandler)

    logger.info("PyQuarantine-Milter starting")
    try:
        # generate milter config
        global_config, config = pyquarantine.generate_milter_config()
    except RuntimeError as e:
        logger.error(e)
        sys.exit(255)

    pyquarantine.QuarantineMilter.global_config = global_config
    pyquarantine.QuarantineMilter.config = config

    # register to have the Milter factory create instances of your class:
    Milter.factory = pyquarantine.QuarantineMilter
    Milter.set_exception_policy(Milter.TEMPFAIL)
    # Milter.set_flags(0)       # tell sendmail which features we use

    # run milter
    rc = 0
    try:
        Milter.runmilter("pyquarantine-milter", socketname=args.socket,
                         timeout=300)
    except Milter.milter.error as e:
        logger.error(e)
        rc = 255
    pyquarantine.mailer.queue.put(None)
    logger.info("PyQuarantine-Milter terminated")
    sys.exit(rc)


if __name__ == "__main__":
    main()



# PyQuarantine-Milter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQuarantine-Milter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQuarantineMilter.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import logging
import os

from calendar import timegm
from datetime import datetime
from glob import glob
from shutil import copyfileobj
from time import gmtime

from pyquarantine import mailer


class BaseQuarantine(object):
    "Quarantine base class"

    def __init__(self, global_config, config, configtest=False):
        self.name = config["name"]
        self.global_config = global_config
        self.config = config
        self.logger = logging.getLogger(__name__)

    def add(self, queueid, mailfrom, recipients, headers,
            fp, subgroups=None, named_subgroups=None):
        "Add email to quarantine."
        fp.seek(0)
        return ""

    def find(self, mailfrom=None, recipients=None, older_than=None):
        "Find emails in quarantine."
        return

    def get_metadata(self, quarantine_id):
        "Return metadata of quarantined email."
        return

    def delete(self, quarantine_id, recipient=None):
        "Delete email from quarantine."
        return

    def notify(self, quarantine_id, recipient=None):
        "Notify recipient about email in quarantine."
        if not self.config["notification_obj"]:
            raise RuntimeError(
                "notification type is set to None, unable to send notifications")
        return

    def release(self, quarantine_id, recipient=None):
        "Release email from quarantine."
        return


class FileQuarantine(BaseQuarantine):
    "Quarantine class to store mails on filesystem."

    def __init__(self, global_config, config, configtest=False):
        super(FileQuarantine, self).__init__(global_config, config, configtest)

        # check if mandatory options are present in config
        for option in ["quarantine_directory"]:
            if option not in self.config.keys() and option in self.global_config.keys():
                self.config[option] = self.global_config[option]
            if option not in self.config.keys():
                raise RuntimeError(
                    "mandatory option '{}' not present in config section '{}' or 'global'".format(
                        option, self.name))
        self.directory = self.config["quarantine_directory"]

        # check if quarantine directory exists and is writable
        if not os.path.isdir(self.directory) or not os.access(
                self.directory, os.W_OK):
            raise RuntimeError(
                "file quarantine directory '{}' does not exist or is not writable".format(
                    self.directory))
        self._metadata_suffix = ".metadata"

    def _save_datafile(self, quarantine_id, fp):
        datafile = os.path.join(self.directory, quarantine_id)
        try:
            with open(datafile, "wb") as f:
                copyfileobj(fp, f)
        except IOError as e:
            raise RuntimeError("unable save data file: {}".format(e))

    def _save_metafile(self, quarantine_id, metadata):
        metafile = os.path.join(
            self.directory, "{}{}".format(
                quarantine_id, self._metadata_suffix))
        try:
            with open(metafile, "w") as f:
                json.dump(metadata, f, indent=2)
        except IOError as e:
            raise RuntimeError("unable to save metadata file: {}".format(e))

    def _remove(self, quarantine_id):
        datafile = os.path.join(self.directory, quarantine_id)
        metafile = "{}{}".format(datafile, self._metadata_suffix)

        try:
            os.remove(metafile)
        except IOError as e:
            raise RuntimeError("unable to remove metadata file: {}".format(e))

        try:
            os.remove(datafile)
        except IOError as e:
            raise RuntimeError("unable to remove data file: {}".format(e))

    def add(self, queueid, mailfrom, recipients, headers,
            fp, subgroups=None, named_subgroups=None):
        "Add email to file quarantine and return quarantine-id."
        super(
            FileQuarantine,
            self).add(
            queueid,
            mailfrom,
            recipients,
            headers,
            fp,
            subgroups,
            named_subgroups)
        quarantine_id = "{}_{}".format(
            datetime.now().strftime("%Y%m%d%H%M%S"), queueid)

        # save mail
        self._save_datafile(quarantine_id, fp)

        # save metadata
        metadata = {
            "mailfrom": mailfrom,
            "recipients": recipients,
            "headers": headers,
            "date": timegm(gmtime()),
            "queue_id": queueid,
            "subgroups": subgroups,
            "named_subgroups": named_subgroups
        }
        try:
            self._save_metafile(quarantine_id, metadata)
        except RuntimeError as e:
            datafile = os.path.join(self.directory, quarantine_id)
            os.remove(datafile)
            raise e

        return quarantine_id

    def get_metadata(self, quarantine_id):
        "Return metadata of quarantined email."
        super(FileQuarantine, self).get_metadata(quarantine_id)

        metafile = os.path.join(
            self.directory, "{}{}".format(
                quarantine_id, self._metadata_suffix))
        if not os.path.isfile(metafile):
            raise RuntimeError(
                "invalid quarantine id '{}'".format(quarantine_id))

        try:
            with open(metafile, "r") as f:
                metadata = json.load(f)
        except IOError as e:
            raise RuntimeError("unable to read metadata file: {}".format(e))
        except json.JSONDecodeError as e:
            raise RuntimeError(
                "invalid meta file '{}': {}".format(
                    metafile, e))

        return metadata

    def find(self, mailfrom=None, recipients=None, older_than=None):
        "Find emails in quarantine."
        super(FileQuarantine, self).find(mailfrom, recipients, older_than)
        if isinstance(mailfrom, str):
            mailfrom = [mailfrom]
        if isinstance(recipients, str):
            recipients = [recipients]

        emails = {}
        metafiles = glob(os.path.join(
            self.directory, "*{}".format(self._metadata_suffix)))
        for metafile in metafiles:
            if not os.path.isfile(metafile):
                continue

            quarantine_id = os.path.basename(
                metafile[:-len(self._metadata_suffix)])
            metadata = self.get_metadata(quarantine_id)
            if older_than is not None:
                if timegm(gmtime()) - metadata["date"] < (older_than * 86400):
                    continue

            if mailfrom is not None:
                if metadata["mailfrom"] not in mailfrom:
                    continue

            if recipients is not None:
                if len(recipients) == 1 and \
                        recipients[0] not in metadata["recipients"]:
                    continue
                elif len(set(recipients + metadata["recipients"])) == len(recipients + metadata["recipients"]):
                    continue

            emails[quarantine_id] = metadata

        return emails

    def delete(self, quarantine_id, recipient=None):
        "Delete email in quarantine."
        super(FileQuarantine, self).delete(quarantine_id, recipient)

        try:
            metadata = self.get_metadata(quarantine_id)
        except RuntimeError as e:
            raise RuntimeError("unable to delete email: {}".format(e))

        if recipient is None:
            self._remove(quarantine_id)
        else:
            if recipient not in metadata["recipients"]:
                raise RuntimeError("invalid recipient '{}'".format(recipient))

            metadata["recipients"].remove(recipient)
            if not metadata["recipients"]:
                self._remove(quarantine_id)
            else:
                self._save_metafile(quarantine_id, metadata)

    def notify(self, quarantine_id, recipient=None):
        "Notify recipient about email in quarantine."
        super(FileQuarantine, self).notify(quarantine_id, recipient)

        try:
            metadata = self.get_metadata(quarantine_id)
        except RuntimeError as e:
            raise RuntimeError("unable to release email: {}".format(e))

        if recipient is not None:
            if recipient not in metadata["recipients"]:
                raise RuntimeError("invalid recipient '{}'".format(recipient))
            recipients = [recipient]
        else:
            recipients = metadata["recipients"]

        datafile = os.path.join(self.directory, quarantine_id)
        try:
            with open(datafile, "rb") as fp:
                self.config["notification_obj"].notify(
                    metadata["queue_id"], quarantine_id, metadata["mailfrom"],
                    recipients, metadata["headers"], fp,
                    metadata["subgroups"], metadata["named_subgroups"],
                    synchronous=True)
        except IOError as e:
            raise RuntimeError

    def release(self, quarantine_id, recipient=None):
        "Release email from quarantine."
        super(FileQuarantine, self).release(quarantine_id, recipient)

        try:
            metadata = self.get_metadata(quarantine_id)
        except RuntimeError as e:
            raise RuntimeError("unable to release email: {}".format(e))

        if recipient is not None:
            if recipient not in metadata["recipients"]:
                raise RuntimeError("invalid recipient '{}'".format(recipient))
            recipients = [recipient]
        else:
            recipients = metadata["recipients"]

        datafile = os.path.join(self.directory, quarantine_id)
        try:
            with open(datafile, "rb") as f:
                mail = f.read()
        except IOError as e:
            raise RuntimeError("unable to read data file: {}".format(e))

        for recipient in recipients:
            try:
                mailer.smtp_send(
                    self.config["smtp_host"],
                    self.config["smtp_port"],
                    metadata["mailfrom"],
                    recipient,
                    mail)
            except Exception as e:
                raise RuntimeError(
                    "error while sending email to '{}': {}".format(
                        recipient, e))

            self.delete(quarantine_id, recipient)


# list of quarantine types and their related quarantine classes
TYPES = {"file": FileQuarantine}



# PyQuarantine-Milter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQuarantine-Milter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQuarantineMilter.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    "QuarantineMilter",
    "generate_milter_config",
    "reload_config",
    "cli",
    "mailer",
    "notifications",
    "quarantines",
    "run",
    "version",
    "whitelists"]

name = "pyquarantine"

import Milter
import configparser
import logging
import os
import re
import sys

from Milter.utils import parse_addr
from collections import defaultdict
from io import BytesIO
from itertools import groupby
from netaddr import IPAddress, IPNetwork
from pyquarantine import quarantines
from pyquarantine import notifications
from pyquarantine import whitelists


class QuarantineMilter(Milter.Base):
    """QuarantineMilter based on Milter.Base to implement milter communication

    The class variable config needs to be filled with the result of the generate_milter_config function.

    """
    config = None
    global_config = None

    # list of default config files
    _config_files = [
        "/etc/pyquarantine/pyquarantine.conf",
        os.path.expanduser('~/pyquarantine.conf'),
        "pyquarantine.conf"]
    # list of possible actions
    _actions = {
        "ACCEPT": Milter.ACCEPT,
        "REJECT": Milter.REJECT,
        "DISCARD": Milter.DISCARD}

    def __init__(self):
        self.logger = logging.getLogger(__name__)
        # save config, it must not change during runtime
        self.global_config = QuarantineMilter.global_config
        self.config = QuarantineMilter.config

    def _get_preferred_quarantine(self):
        matching_quarantines = [
            q for q in self.recipients_quarantines.values() if q]
        if self.global_config["preferred_quarantine_action"] == "first":
            quarantine = sorted(
                matching_quarantines,
                key=lambda x: x["index"])[0]
        else:
            quarantine = sorted(
                matching_quarantines,
                key=lambda x: x["index"],
                reverse=True)[0]
        return quarantine

    @staticmethod
    def get_configfiles():
        return QuarantineMilter._config_files

    @staticmethod
    def get_actions():
        return QuarantineMilter._actions

    @staticmethod
    def set_configfiles(config_files):
        QuarantineMilter._config_files = config_files

    def connect(self, IPname, family, hostaddr):
        self.logger.debug(
            "accepted milter connection from {} port {}".format(
                *hostaddr))
        ip = IPAddress(hostaddr[0])
        for quarantine in self.config.copy():
            for ignore in quarantine["ignore_hosts_list"]:
                if ip in ignore:
                    self.logger.debug(
                        "host {} is ignored by quarantine {}".format(
                            hostaddr[0], quarantine["name"]))
                    self.config.remove(quarantine)
                    break
            if not self.config:
                self.logger.debug(
                    "host {} is ignored by all quarantines, "
                    "skip further processing",
                    hostaddr[0])
                return Milter.ACCEPT
        return Milter.CONTINUE

    @Milter.noreply
    def envfrom(self, mailfrom, *str):
        self.mailfrom = "@".join(parse_addr(mailfrom)).lower()
        self.recipients = set()
        return Milter.CONTINUE

    @Milter.noreply
    def envrcpt(self, to, *str):
        self.recipients.add("@".join(parse_addr(to)).lower())
        return Milter.CONTINUE

    @Milter.noreply
    def data(self):
        self.queueid = self.getsymval('i')
        self.logger.debug(
            "{}: received queue-id from MTA".format(self.queueid))
        self.recipients = list(self.recipients)
        self.headers = []
        return Milter.CONTINUE

    @Milter.noreply
    def header(self, name, value):
        self.headers.append((name, value))
        return Milter.CONTINUE

    def eoh(self):
        try:
            self.whitelist_cache = whitelists.WhitelistCache()

            # initialize dicts to set quaranines per recipient and keep matches
            self.recipients_quarantines = {}
            self.quarantines_matches = {}

            # iterate email headers
            recipients_to_check = self.recipients.copy()
            for name, value in self.headers:
                header = "{}: {}".format(name, value)
                self.logger.debug(
                    "{}: checking header against configured quarantines: {}".format(
                        self.queueid, header))
                # iterate quarantines
                for quarantine in self.config:
                    if len(self.recipients_quarantines) == len(
                            self.recipients):
                        # every recipient matched a quarantine already
                        if quarantine["index"] >= max(
                                [q["index"] for q in self.recipients_quarantines.values()]):
                            # all recipients matched a quarantine with at least
                            # the same precedence already, skip checks against
                            # quarantines with lower precedence
                            self.logger.debug(
                                "{}: {}: skip further checks of this header".format(
                                    self.queueid, quarantine["name"]))
                            break

                    # check email header against quarantine regex
                    self.logger.debug(
                        "{}: {}: checking header against regex '{}'".format(
                            self.queueid, quarantine["name"], quarantine["regex"]))
                    match = quarantine["regex_compiled"].search(header)
                    if match:
                        self.logger.debug(
                            "{}: {}: header matched regex".format(
                                self.queueid, quarantine["name"]))
                        # check for whitelisted recipients
                        whitelist = quarantine["whitelist_obj"]
                        if whitelist is not None:
                            try:
                                whitelisted_recipients = self.whitelist_cache.get_whitelisted_recipients(
                                    whitelist, self.mailfrom, recipients_to_check)
                            except RuntimeError as e:
                                self.logger.error(
                                    "{}: {}: unable to query whitelist: {}".format(
                                        self.queueid, quarantine["name"], e))
                                return Milter.TEMPFAIL
                        else:
                            whitelisted_recipients = {}

                        # iterate recipients
                        for recipient in recipients_to_check.copy():
                            if recipient in whitelisted_recipients:
                                # recipient is whitelisted in this quarantine
                                self.logger.debug(
                                    "{}: {}: recipient '{}' is whitelisted".format(
                                        self.queueid, quarantine["name"], recipient))
                                continue

                            if recipient not in self.recipients_quarantines.keys() or \
                                    self.recipients_quarantines[recipient]["index"] > quarantine["index"]:
                                self.logger.debug(
                                    "{}: {}: set quarantine for recipient '{}'".format(
                                        self.queueid, quarantine["name"], recipient))
                                # save match for later use as template
                                # variables
                                self.quarantines_matches[quarantine["name"]] = match
                                self.recipients_quarantines[recipient] = quarantine
                                if quarantine["index"] == 0:
                                    # we do not need to check recipients which
                                    # matched the quarantine with the highest
                                    # precedence already
                                    recipients_to_check.remove(recipient)
                            else:
                                self.logger.debug(
                                    "{}: {}: a quarantine with same or higher precedence "
                                    "matched already for recipient '{}'".format(
                                        self.queueid, quarantine["name"], recipient))

                if not recipients_to_check:
                    self.logger.debug(
                        "{}: all recipients matched the first quarantine, "
                        "skipping all remaining header checks".format(
                            self.queueid))
                    break

            # check if no quarantine has matched for all recipients
            if not self.recipients_quarantines:
                # accept email
                self.logger.info(
                    "{}: passed clean for all recipients".format(
                        self.queueid))
                return Milter.ACCEPT

            # check if the email body is needed
            keep_body = False
            for recipient, quarantine in self.recipients_quarantines.items():
                if quarantine["quarantine_obj"] or quarantine["notification_obj"]:
                    keep_body = True
                    break

            if keep_body:
                self.logger.debug(
                    "{}: initializing memory buffer to save email data".format(
                        self.queueid))
                # initialize memory buffer to save email data
                self.fp = BytesIO()
                # write email headers to memory buffer
                for name, value in self.headers:
                    self.fp.write("{}: {}\n".format(name, value).encode())
                self.fp.write("\n".encode())
            else:
                # quarantine and notification are disabled on all matching
                # quarantines, return configured action
                quarantine = self._get_preferred_quarantine()
                self.logger.info(
                    "{}: {} matching quarantine is '{}', performing milter action {}".format(
                        self.queueid,
                        self.global_config["preferred_quarantine_action"],
                        quarantine["name"],
                        quarantine["action"].upper()))
                if quarantine["action"] == "reject":
                    self.setreply("554", "5.7.0", quarantine["reject_reason"])
                return quarantine["milter_action"]

            return Milter.CONTINUE

        except Exception as e:
            self.logger.exception(
                "an exception occured in eoh function: {}".format(e))
            return Milter.TEMPFAIL

    def body(self, chunk):
        try:
            # save received body chunk
            self.fp.write(chunk)
        except Exception as e:
            self.logger.exception(
                "an exception occured in body function: {}".format(e))
            return Milter.TEMPFAIL
        return Milter.CONTINUE

    def eom(self):
        try:
            # processing recipients grouped by quarantines
            quarantines = []
            for quarantine, recipients in groupby(
                    sorted(self.recipients_quarantines,
                           key=lambda x: self.recipients_quarantines[x]["index"]),
                    lambda x: self.recipients_quarantines[x]):
                quarantines.append((quarantine, list(recipients)))

            # iterate quarantines sorted by index
            for quarantine, recipients in sorted(
                    quarantines, key=lambda x: x[0]["index"]):
                quarantine_id = ""
                headers = defaultdict(str)
                for name, value in self.headers:
                    headers[name.lower()] = value
                subgroups = self.quarantines_matches[quarantine["name"]].groups(
                    default="")
                named_subgroups = self.quarantines_matches[quarantine["name"]].groupdict(
                    default="")

                # check if a quarantine is configured
                if quarantine["quarantine_obj"] is not None:
                    # add email to quarantine
                    self.logger.info("{}: adding to quarantine '{}' for: {}".format(
                        self.queueid, quarantine["name"], ", ".join(recipients)))
                    try:
                        quarantine_id = quarantine["quarantine_obj"].add(
                            self.queueid, self.mailfrom, recipients, headers, self.fp,
                            subgroups, named_subgroups)
                    except RuntimeError as e:
                        self.logger.error(
                            "{}: unable to add to quarantine '{}': {}".format(
                                self.queueid, quarantine["name"], e))
                        return Milter.TEMPFAIL

                # check if a notification is configured
                if quarantine["notification_obj"] is not None:
                    # notify
                    self.logger.info(
                        "{}: sending notification for quarantine '{}' to: {}".format(
                            self.queueid, quarantine["name"], ", ".join(recipients)))
                    try:
                        quarantine["notification_obj"].notify(
                            self.queueid, quarantine_id,
                            self.mailfrom, recipients, headers, self.fp,
                            subgroups, named_subgroups)
                    except RuntimeError as e:
                        self.logger.error(
                            "{}: unable to send notification for quarantine '{}': {}".format(
                                self.queueid, quarantine["name"], e))
                        return Milter.TEMPFAIL

                # remove processed recipient
                for recipient in recipients:
                    self.delrcpt(recipient)
                    self.recipients.remove(recipient)

            self.fp.close()

            # email passed clean for at least one recipient, accepting email
            if self.recipients:
                self.logger.info(
                    "{}: passed clean for: {}".format(
                        self.queueid, ", ".join(
                            self.recipients)))
                return Milter.ACCEPT

            # return configured action
            quarantine = self._get_preferred_quarantine()
            self.logger.info(
                "{}: {} matching quarantine is '{}', performing milter action {}".format(
                    self.queueid,
                    self.global_config["preferred_quarantine_action"],
                    quarantine["name"],
                    quarantine["action"].upper()))
            if quarantine["action"] == "reject":
                self.setreply("554", "5.7.0", quarantine["reject_reason"])
            return quarantine["milter_action"]

        except Exception as e:
            self.logger.exception(
                "an exception occured in eom function: {}".format(e))
            return Milter.TEMPFAIL


def generate_milter_config(configtest=False, config_files=[]):
    "Generate the configuration for QuarantineMilter class."
    logger = logging.getLogger(__name__)

    # read config file
    parser = configparser.ConfigParser()
    if not config_files:
        config_files = parser.read(QuarantineMilter.get_configfiles())
    else:
        config_files = parser.read(config_files)
    if not config_files:
        raise RuntimeError("config file not found")

    QuarantineMilter.set_configfiles(config_files)
    os.chdir(os.path.dirname(config_files[0]))

    # check if mandatory config options in global section are present
    if "global" not in parser.sections():
        raise RuntimeError(
            "mandatory section 'global' not present in config file")
    for option in ["quarantines", "preferred_quarantine_action"]:
        if not parser.has_option("global", option):
            raise RuntimeError(
                "mandatory option '{}' not present in config section 'global'".format(option))

    # read global config section
    global_config = dict(parser.items("global"))
    global_config["preferred_quarantine_action"] = global_config["preferred_quarantine_action"].lower()
    if global_config["preferred_quarantine_action"] not in ["first", "last"]:
        raise RuntimeError(
            "option preferred_quarantine_action has illegal value")

    # read active quarantine names
    quarantine_names = [
        q.strip() for q in global_config["quarantines"].split(",")]
    if len(quarantine_names) != len(set(quarantine_names)):
        raise RuntimeError(
            "at least one quarantine is specified multiple times in quarantines option")
    if "global" in quarantine_names:
        quarantine_names.remove("global")
        logger.warning(
            "removed illegal quarantine name 'global' from list of active quarantines")
    if not quarantine_names:
        raise RuntimeError("no quarantines configured")

    milter_config = []

    logger.debug("preparing milter configuration ...")
    # iterate quarantine names
    for index, quarantine_name in enumerate(quarantine_names):

        # check if config section for current quarantine exists
        if quarantine_name not in parser.sections():
            raise RuntimeError(
                "config section '{}' does not exist".format(quarantine_name))
        config = dict(parser.items(quarantine_name))

        # check if mandatory config options are present in config
        for option in ["regex", "quarantine_type", "notification_type",
                       "action", "whitelist_type", "smtp_host", "smtp_port"]:
            if option not in config.keys() and \
                    option in global_config.keys():
                config[option] = global_config[option]
            if option not in config.keys():
                raise RuntimeError(
                    "mandatory option '{}' not present in config section '{}' or 'global'".format(
                        option, quarantine_name))

        # check if optional config options are present in config
        defaults = {
            "reject_reason": "Message rejected",
            "ignore_hosts": ""
        }
        for option in defaults.keys():
            if option not in config.keys() and \
                    option in global_config.keys():
                config[option] = global_config[option]
            if option not in config.keys():
                config[option] = defaults[option]

        # set quarantine name
        config["name"] = quarantine_name

        # set the index
        config["index"] = index

        # pre-compile regex
        logger.debug(
            "{}: compiling regex '{}'".format(
                quarantine_name,
                config["regex"]))
        config["regex_compiled"] = re.compile(
            config["regex"], re.MULTILINE + re.DOTALL + re.IGNORECASE)

        # create quarantine instance
        quarantine_type = config["quarantine_type"].lower()
        if quarantine_type in quarantines.TYPES.keys():
            logger.debug(
                "{}: initializing quarantine type '{}'".format(
                    quarantine_name,
                    quarantine_type.upper()))
            quarantine = quarantines.TYPES[quarantine_type](
                global_config, config, configtest)
        elif quarantine_type == "none":
            logger.debug("{}: quarantine is NONE".format(quarantine_name))
            quarantine = None
        else:
            raise RuntimeError(
                "{}: unknown quarantine type '{}'".format(
                    quarantine_name, quarantine_type))

        config["quarantine_obj"] = quarantine

        # create whitelist instance
        whitelist_type = config["whitelist_type"].lower()
        if whitelist_type in whitelists.TYPES.keys():
            logger.debug(
                "{}: initializing whitelist type '{}'".format(
                    quarantine_name,
                    whitelist_type.upper()))
            whitelist = whitelists.TYPES[whitelist_type](
                global_config, config, configtest)
        elif whitelist_type == "none":
            logger.debug("{}: whitelist is NONE".format(quarantine_name))
            whitelist = None
        else:
            raise RuntimeError(
                "{}: unknown whitelist type '{}'".format(
                    quarantine_name, whitelist_type))

        config["whitelist_obj"] = whitelist

        # create notification instance
        notification_type = config["notification_type"].lower()
        if notification_type in notifications.TYPES.keys():
            logger.debug(
                "{}: initializing notification type '{}'".format(
                    quarantine_name,
                    notification_type.upper()))
            notification = notifications.TYPES[notification_type](
                global_config, config, configtest)
        elif notification_type == "none":
            logger.debug("{}: notification is NONE".format(quarantine_name))
            notification = None
        else:
            raise RuntimeError(
                "{}: unknown notification type '{}'".format(
                    quarantine_name, notification_type))

        config["notification_obj"] = notification

        # determining milter action for this quarantine
        action = config["action"].upper()
        if action in QuarantineMilter.get_actions().keys():
            logger.debug("{}: action is {}".format(quarantine_name, action))
            config["milter_action"] = QuarantineMilter.get_actions()[action]
        else:
            raise RuntimeError(
                "{}: unknown action '{}'".format(
                    quarantine_name, action))

        # create host/network whitelist
        config["ignore_hosts_list"] = []
        ignored = set([p.strip()
                       for p in config["ignore_hosts"].split(",") if p])
        for ignore in ignored:
            if not ignore:
                continue
            # parse network notation
            try:
                net = IPNetwork(ignore)
            except AddrFormatError as e:
                raise RuntimeError("error parsing ignore_hosts: {}".format(e))
            else:
                config["ignore_hosts_list"].append(net)
        if config["ignore_hosts_list"]:
            logger.debug(
                "{}: ignore hosts: {}".format(
                    quarantine_name,
                    ", ".join(ignored)))

        milter_config.append(config)

    return global_config, milter_config


def reload_config():
    "Reload the configuration of QuarantineMilter class."
    logger = logging.getLogger(__name__)

    try:
        global_config, config = generate_milter_config()
    except RuntimeError as e:
        logger.info(e)
        logger.info("daemon is still running with previous configuration")
    else:
        logger.info("reloading configuration")
        QuarantineMilter.global_config = global_config
        QuarantineMilter.config = config



# PyQuarantine-Milter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQuarantine-Milter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQuarantineMilter.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import peewee
import re
import sys

from datetime import datetime
from playhouse.db_url import connect


class WhitelistBase(object):
    "Whitelist base class"

    def __init__(self, global_config, config, configtest=False):
        self.global_config = global_config
        self.config = config
        self.configtest = configtest
        self.name = config["name"]
        self.logger = logging.getLogger(__name__)
        self.valid_entry_regex = re.compile(
            r"^[a-zA-Z0-9_.+-]*?(@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)?$")

    def check(self, mailfrom, recipient):
        "Check if mailfrom/recipient combination is whitelisted."
        return

    def find(self, mailfrom=None, recipients=None, older_than=None):
        "Find whitelist entries."
        return

    def add(self, mailfrom, recipient, comment, permanent):
        "Add entry to whitelist."
        # check if mailfrom and recipient are valid
        if not self.valid_entry_regex.match(mailfrom):
            raise RuntimeError("invalid from address")
        if not self.valid_entry_regex.match(recipient):
            raise RuntimeError("invalid recipient")
        return

    def delete(self, whitelist_id):
        "Delete entry from whitelist."
        return


class WhitelistModel(peewee.Model):
    mailfrom = peewee.CharField()
    recipient = peewee.CharField()
    created = peewee.DateTimeField(default=datetime.now)
    last_used = peewee.DateTimeField(default=datetime.now)
    comment = peewee.TextField(default="")
    permanent = peewee.BooleanField(default=False)


class Meta(object):
    indexes = (
        # trailing comma is mandatory if only one index should be created
        (('mailfrom', 'recipient'), True),
    )


class DatabaseWhitelist(WhitelistBase):
    "Whitelist class to store whitelist in a database"
    _db_connections = {}
    _db_tables = {}

    def __init__(self, global_config, config, configtest=False):
        super(
            DatabaseWhitelist,
            self).__init__(
            global_config,
            config,
            configtest)

        # check if mandatory options are present in config
        for option in ["whitelist_db_connection", "whitelist_db_table"]:
            if option not in self.config.keys() and option in self.global_config.keys():
                self.config[option] = self.global_config[option]
            if option not in self.config.keys():
                raise RuntimeError(
                    "mandatory option '{}' not present in config section '{}' or 'global'".format(
                        option, self.name))

        tablename = self.config["whitelist_db_table"]
        connection_string = self.config["whitelist_db_connection"]

        if connection_string in DatabaseWhitelist._db_connections.keys():
            db = DatabaseWhitelist._db_connections[connection_string]
        else:
            try:
                # connect to database
                self.logger.debug(
                    "connecting to database '{}'".format(
                        re.sub(
                            r"(.*?://.*?):.*?(@.*)",
                            r"\1:<PASSWORD>\2",
                            connection_string)))
                db = connect(connection_string)
            except Exception as e:
                raise RuntimeError(
                    "unable to connect to database: {}".format(e))

            DatabaseWhitelist._db_connections[connection_string] = db

        # generate model meta class
        self.meta = Meta
        self.meta.database = db
        self.meta.table_name = tablename
        self.model = type("WhitelistModel_{}".format(self.name), (WhitelistModel,), {
            "Meta": self.meta
        })

        if connection_string not in DatabaseWhitelist._db_tables.keys():
            DatabaseWhitelist._db_tables[connection_string] = []

        if tablename not in DatabaseWhitelist._db_tables[connection_string]:
            DatabaseWhitelist._db_tables[connection_string].append(tablename)
            if not self.configtest:
                try:
                    db.create_tables([self.model])
                except Exception as e:
                    raise RuntimeError(
                        "unable to initialize table '{}': {}".format(
                            tablename, e))

    def _entry_to_dict(self, entry):
        result = {}
        result[entry.id] = {
            "id": entry.id,
            "mailfrom": entry.mailfrom,
            "recipient": entry.recipient,
            "created": entry.created,
            "last_used": entry.last_used,
            "comment": entry.comment,
            "permanent": entry.permanent
        }
        return result

    def get_weight(self, entry):
        value = 0
        for address in [entry.mailfrom, entry.recipient]:
            if address == "":
                value += 2
            elif address[0] == "@":
                value += 1
        return value

    def check(self, mailfrom, recipient):
        # check if mailfrom/recipient combination is whitelisted
        super(DatabaseWhitelist, self).check(mailfrom, recipient)

        # generate list of possible mailfroms
        self.logger.debug(
            "query database for whitelist entries from <{}> to <{}>".format(
                mailfrom, recipient))
        mailfroms = [""]
        if "@" in mailfrom and not mailfrom.startswith("@"):
            mailfroms.append("@{}".format(mailfrom.split("@")[1]))
        mailfroms.append(mailfrom)

        # generate list of possible recipients
        recipients = [""]
        if "@" in recipient and not recipient.startswith("@"):
            recipients.append("@{}".format(recipient.split("@")[1]))
        recipients.append(recipient)

        # query the database
        try:
            entries = list(
                self.model.select().where(
                    self.model.mailfrom.in_(mailfroms),
                    self.model.recipient.in_(recipients)))
        except Exception as e:
            raise RuntimeError("unable to query database: {}".format(e))

        if not entries:
            # no whitelist entry found
            return {}

        if len(entries) > 1:
            entries.sort(key=lambda x: self.get_weight(x), reverse=True)

        # use entry with the highest weight
        entry = entries[0]
        entry.last_used = datetime.now()
        entry.save()
        result = {}
        for entry in entries:
            result.update(self._entry_to_dict(entry))

        return result

    def find(self, mailfrom=None, recipients=None, older_than=None):
        "Find whitelist entries."
        super(DatabaseWhitelist, self).find(mailfrom, recipients, older_than)

        if isinstance(mailfrom, str):
            mailfrom = [mailfrom]
        if isinstance(recipients, str):
            recipients = [recipients]

        entries = {}
        try:
            for entry in list(self.model.select()):
                if older_than is not None:
                    if (datetime.now() - entry.last_used).total_seconds() < (older_than * 86400):
                        continue

                if mailfrom is not None:
                    if entry.mailfrom not in mailfrom:
                        continue

                if recipients is not None:
                    if entry.recipient not in recipients:
                        continue

                entries.update(self._entry_to_dict(entry))
        except Exception as e:
            raise RuntimeError("unable to query database: {}".format(e))

        return entries

    def add(self, mailfrom, recipient, comment, permanent):
        "Add entry to whitelist."
        super(
            DatabaseWhitelist,
            self).add(
            mailfrom,
            recipient,
            comment,
            permanent)

        try:
            self.model.create(
                mailfrom=mailfrom,
                recipient=recipient,
                comment=comment,
                permanent=permanent)
        except Exception as e:
            raise RuntimeError("unable to add entry to database: {}".format(e))

    def delete(self, whitelist_id):
        "Delete entry from whitelist."
        super(DatabaseWhitelist, self).delete(whitelist_id)

        try:
            query = self.model.delete().where(self.model.id == whitelist_id)
            deleted = query.execute()
        except Exception as e:
            raise RuntimeError(
                "unable to delete entry from database: {}".format(e))

        if deleted == 0:
            raise RuntimeError("invalid whitelist id")


class WhitelistCache(object):
    def __init__(self):
        self.cache = {}

    def load(self, whitelist, mailfrom, recipients):
        for recipient in recipients:
            self.check(whitelist, mailfrom, recipient)

    def check(self, whitelist, mailfrom, recipient):
        if whitelist not in self.cache.keys():
            self.cache[whitelist] = {}
        if recipient not in self.cache[whitelist].keys():
            self.cache[whitelist][recipient] = None
        if self.cache[whitelist][recipient] is None:
            self.cache[whitelist][recipient] = whitelist.check(
                mailfrom, recipient)
        return self.cache[whitelist][recipient]

    def get_whitelisted_recipients(self, whitelist, mailfrom, recipients):
        self.load(whitelist, mailfrom, recipients)
        return list(
            filter(lambda x: self.cache[whitelist][x], self.cache[whitelist].keys()))


# list of whitelist types and their related whitelist classes
TYPES = {"db": DatabaseWhitelist}



# PyQuarantine-Milter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQuarantine-Milter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQuarantineMilter.  If not, see <http://www.gnu.org/licenses/>.
#

import email
import logging
import re

from bs4 import BeautifulSoup
from cgi import escape
from collections import defaultdict
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from os.path import basename

from pyquarantine import mailer


class BaseNotification(object):
    "Notification base class"

    def __init__(self, global_config, config, configtest=False):
        self.quarantine_name = config["name"]
        self.global_config = global_config
        self.config = config
        self.logger = logging.getLogger(__name__)

    def notify(self, queueid, quarantine_id, mailfrom, recipients, headers,
               fp, subgroups=None, named_subgroups=None, synchronous=False):
        fp.seek(0)
        pass


class EMailNotification(BaseNotification):
    "Notification class to send notifications via mail."
    _html_text = "text/html"
    _plain_text = "text/plain"
    _bad_tags = [
        "applet",
        "embed",
        "frame",
        "frameset",
        "head",
        "iframe",
        "script"
    ]
    _good_tags = [
        "a",
        "b",
        "br",
        "center",
        "div",
        "font",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "i",
        "img",
        "li",
        "p",
        "pre",
        "span",
        "table",
        "td",
        "th",
        "tr",
        "tt",
        "u",
        "ul"
    ]
    _good_attributes = [
        "align",
        "alt",
        "bgcolor",
        "border",
        "cellpadding",
        "cellspacing",
        "class",
        "color",
        "colspan",
        "dir",
        "face",
        "headers",
        "height",
        "id",
        "name",
        "rowspan",
        "size",
        "src",
        "style",
        "title",
        "type",
        "valign",
        "value",
        "width"
    ]

    def __init__(self, global_config, config, configtest=False):
        super(EMailNotification, self).__init__(
            global_config, config, configtest)

        # check if mandatory options are present in config
        for option in [
            "notification_email_smtp_host",
            "notification_email_smtp_port",
            "notification_email_envelope_from",
            "notification_email_from",
            "notification_email_subject",
            "notification_email_template",
                "notification_email_embedded_imgs"]:
            if option not in self.config.keys() and option in self.global_config.keys():
                self.config[option] = self.global_config[option]
            if option not in self.config.keys():
                raise RuntimeError(
                    "mandatory option '{}' not present in config section '{}' or 'global'".format(
                        option, self.quarantine_name))

        # check if optional config options are present in config
        defaults = {
            "notification_email_replacement_img": "",
            "notification_email_strip_images": "false",
            "notification_email_parser_lib": "lxml"
        }
        for option in defaults.keys():
            if option not in config.keys() and \
                    option in global_config.keys():
                config[option] = global_config[option]
            if option not in config.keys():
                config[option] = defaults[option]

        self.smtp_host = self.config["notification_email_smtp_host"]
        self.smtp_port = self.config["notification_email_smtp_port"]
        self.mailfrom = self.config["notification_email_envelope_from"]
        self.from_header = self.config["notification_email_from"]
        self.subject = self.config["notification_email_subject"]

        testvars = defaultdict(str, test="TEST")

        # test-parse from header
        try:
            self.from_header.format_map(testvars)
        except ValueError as e:
            raise RuntimeError(
                "error parsing notification_email_from: {}".format(e))

        # test-parse subject
        try:
            self.subject.format_map(testvars)
        except ValueError as e:
            raise RuntimeError(
                "error parsing notification_email_subject: {}".format(e))

        # read and parse email notification template
        try:
            self.template = open(
                self.config["notification_email_template"], "r").read()
            self.template.format_map(testvars)
        except IOError as e:
            raise RuntimeError("error reading template: {}".format(e))
        except ValueError as e:
            raise RuntimeError("error parsing template: {}".format(e))

        strip_images = self.config["notification_email_strip_images"].strip().upper()
        if strip_images in ["TRUE", "ON", "YES"]:
            self.strip_images = True
        elif strip_images in ["FALSE", "OFF", "NO"]:
            self.strip_images = False
        else:
            raise RuntimeError("error parsing notification_email_strip_images: unknown value")

        self.parser_lib = self.config["notification_email_parser_lib"].strip()
        if self.parser_lib not in ["lxml", "html.parser"]:
            raise RuntimeError("error parsing notification_email_parser_lib: unknown value")

        # read email replacement image if specified
        replacement_img = self.config["notification_email_replacement_img"].strip()
        if not self.strip_images and replacement_img:
            try:
                self.replacement_img = MIMEImage(
                    open(replacement_img, "rb").read())
            except IOError as e:
                raise RuntimeError(
                    "error reading replacement image: {}".format(e))
            else:
                self.replacement_img.add_header(
                    "Content-ID", "<removed_for_security_reasons>")
        else:
            self.replacement_img = None

        # read images to embed if specified
        embedded_img_paths = [
            p.strip() for p in self.config["notification_email_embedded_imgs"].split(",") if p]
        self.embedded_imgs = []
        for img_path in embedded_img_paths:
            # read image
            try:
                img = MIMEImage(open(img_path, "rb").read())
            except IOError as e:
                raise RuntimeError("error reading image: {}".format(e))
            else:
                img.add_header("Content-ID", "<{}>".format(basename(img_path)))
                self.embedded_imgs.append(img)

    def get_text(self, queueid, part):
        "Get the mail text in html form from email part."
        mimetype = part.get_content_type()

        self.logger.debug(
            "{}: extracting content of email text part".format(queueid))
        text = part.get_payload(decode=True)

        if mimetype == EMailNotification._plain_text:
            self.logger.debug(
                "{}: content mimetype is {}, converting to {}".format(
                    queueid, mimetype, self._html_text))
            text = re.sub(r"^(.*)$", r"\1<br/>",
                          escape(text.decode()), flags=re.MULTILINE)
        else:
            self.logger.debug(
                "{}: content mimetype is {}".format(
                    queueid, mimetype))
        self.logger.debug(
            "{}: trying to create BeatufilSoup object with parser lib {}, "
            "text length is {} bytes".format(
                queueid, self.parser_lib, len(text)))
        soup = BeautifulSoup(text, self.parser_lib)
        self.logger.debug(
            "{}: sucessfully created BeautifulSoup object".format(queueid))
        return soup

    def get_text_multipart(self, queueid, msg, preferred=_html_text):
        "Get the mail text of a multipart email in html form."
        soup = None

        for part in msg.get_payload():
            mimetype = part.get_content_type()
            if mimetype in [EMailNotification._plain_text,
                            EMailNotification._html_text]:
                soup = self.get_text(queueid, part)
            elif mimetype.startswith("multipart"):
                soup = self.get_text_multipart(queueid, part, preferred)

            if soup is not None and mimetype == preferred:
                break
        return soup

    def sanitize(self, queueid, soup):
        "Sanitize mail html text."
        self.logger.debug("{}: sanitizing email text".format(queueid))

        # completly remove bad elements
        for element in soup(EMailNotification._bad_tags):
            self.logger.debug(
                "{}: removing dangerous tag '{}' and its content".format(
                    queueid, element.name))
            element.extract()

        # remove not whitelisted elements, but keep their content
        for element in soup.find_all(True):
            if element.name not in EMailNotification._good_tags:
                self.logger.debug(
                    "{}: removing tag '{}', keep its content".format(
                        queueid, element.name))
                element.replaceWithChildren()

        # remove not whitelisted attributes
        for element in soup.find_all(True):
            for attribute in list(element.attrs.keys()):
                if attribute not in EMailNotification._good_attributes:
                    if element.name == "a" and attribute == "href":
                        self.logger.debug(
                            "{}: setting attribute href to '#' on tag '{}'".format(
                                queueid, element.name))
                        element["href"] = "#"
                    else:
                        self.logger.debug(
                            "{}: removing attribute '{}' from tag '{}'".format(
                                queueid, attribute, element.name))
                        del(element.attrs[attribute])
        return soup

    def get_html_text_part(self, queueid, msg):
        "Get the mail text of an email in html form."
        soup = None
        mimetype = msg.get_content_type()

        self.logger.debug(
            "{}: trying to find text part of email".format(queueid))
        if mimetype in [EMailNotification._plain_text,
                        EMailNotification._html_text]:
            soup = self.get_text(queueid, msg)
        elif mimetype.startswith("multipart"):
            soup = self.get_text_multipart(queueid, msg)

        if soup is None:
            self.logger.error(
                "{}: unable to extract text part of email".format(queueid))
            text = "ERROR: unable to extract text from email body"
            soup = BeautifulSoup(text, "lxml", "UTF-8")

        return soup

    def notify(self, queueid, quarantine_id, mailfrom, recipients, headers, fp,
               subgroups=None, named_subgroups=None, synchronous=False):
        "Notify recipients via email."
        super(
            EMailNotification,
            self).notify(
            queueid,
            quarantine_id,
            mailfrom,
            recipients,
            headers,
            fp,
            subgroups,
            named_subgroups,
            synchronous)

        # extract html text from email
        self.logger.debug(
            "{}: extraction email text from original email".format(queueid))
        soup = self.get_html_text_part(
            queueid, email.message_from_binary_file(fp))

        # replace picture sources
        image_replaced = False
        if self.strip_images:
            self.logger.debug(
                "{}: looking for images to strip".format(queueid))
            for element in soup("img"):
                if "src" in element.attrs.keys():
                    self.logger.debug(
                        "{}: strip image: {}".format(
                            queueid, element["src"]))
                element.extract()
        elif self.replacement_img:
            self.logger.debug(
                "{}: looking for images to replace".format(queueid))
            for element in soup("img"):
                if "src" in element.attrs.keys():
                    self.logger.debug(
                        "{}: replacing image: {}".format(
                            queueid, element["src"]))
                element["src"] = "cid:removed_for_security_reasons"
                image_replaced = True

        # sanitizing email text of original email
        sanitized_text = self.sanitize(queueid, soup)
        del soup

        # sending email notifications
        for recipient in recipients:
            self.logger.debug(
                "{}: generating notification email for '{}'".format(
                    queueid, recipient))
            self.logger.debug("{}: parsing email template".format(queueid))

            # generate dict containing all template variables
            variables = defaultdict(str,
                                    EMAIL_HTML_TEXT=sanitized_text,
                                    EMAIL_FROM=escape(headers["from"]),
                                    EMAIL_ENVELOPE_FROM=escape(mailfrom),
                                    EMAIL_TO=escape(recipient),
                                    EMAIL_SUBJECT=escape(headers["subject"]),
                                    EMAIL_QUARANTINE_ID=quarantine_id)

            if subgroups:
                number = 0
                for subgroup in subgroups:
                    variables["SUBGROUP_{}".format(number)] = escape(subgroup)
            if named_subgroups:
                for key, value in named_subgroups.items():
                    named_subgroups[key] = escape(value)
                variables.update(named_subgroups)

            # parse template
            htmltext = self.template.format_map(variables)

            msg = MIMEMultipart('related')
            msg["From"] = self.from_header.format_map(
                defaultdict(str, EMAIL_FROM=headers["from"]))
            msg["To"] = headers["to"]
            msg["Subject"] = self.subject.format_map(
                defaultdict(str, EMAIL_SUBJECT=headers["subject"]))
            msg["Date"] = email.utils.formatdate()
            msg.attach(MIMEText(htmltext, "html", 'UTF-8'))

            if image_replaced:
                self.logger.debug(
                    "{}: attaching notification_replacement_img".format(queueid))
                msg.attach(self.replacement_img)

            for img in self.embedded_imgs:
                self.logger.debug("{}: attaching imgage".format(queueid))
                msg.attach(img)

            self.logger.debug(
                "{}: sending notification email to: {}".format(
                    queueid, recipient))
            if synchronous:
                try:
                    mailer.smtp_send(self.smtp_host, self.smtp_port,
                                     self.mailfrom, recipient, msg.as_string())
                except Exception as e:
                    raise RuntimeError(
                        "error while sending email to '{}': {}".format(
                            recipient, e))
            else:
                mailer.sendmail(self.smtp_host, self.smtp_port, queueid,
                                self.mailfrom, recipient, msg.as_string(),
                                "notification email")


# list of notification types and their related notification classes
TYPES = {"email": EMailNotification}
